/************************************************************************/
/*									*/
/*  Procedures to read an ispell affix file and dictionary files and	*/
/*  to build a finite automaton that recognises the same words.		*/
/*									*/
/*  Treatment of ispell affix files.					*/
/*  Syntax is simplified. (No line breaks allowed in rule.)		*/
/*									*/
/************************************************************************/

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<string.h>
#   include	<ctype.h>
#   include	<charnames.h>
#   include	<ind.h>
#   include	<uniUtf8.h>
#   include	<ucd.h>
#   include	<appDebugon.h>
#   include	"affix.h"

# define BUL 400

/************************************************************************/
/*									*/
/*  Read one affix character.						*/
/*									*/
/************************************************************************/

int affixChar(		const AffixRules *	ars,
			const char *		buf,
			iconv_t			icv,
			int			lower,
			int *			pPos )
    {
    int			pos= *pPos;
    int			i;

    unsigned short	w[2];

    for ( i= 0; i < ars->arStringCharCount; i++ )
	{
	const char *	str= (char *)ars->arStringChars[i].scString;
	int		len= ars->arStringChars[i].scLength;

	if  ( ! strncmp( buf+ pos, str, len ) )
	    {
	    w[0]= ars->arStringChars[i].scMeaningUtf16;

	    if  ( lower && ucdIsLu( w[0] ) )
		{ w[0]= ucdToLower( w[0] );	}

	    *pPos += len;
	    return w[0];
	    }
	}

    if  ( affixToUtf16( icv, w, 1, buf+ pos, 1 ) )
	{ LDEB(1);	}

    (*pPos)++;

    if  ( utilIndexSetContains( &(ars->arWordChars), buf[pos]&0xff ) )
	{
	if  ( lower && ucdIsLu( w[0] ) )
	    { w[0]= ucdToLower( w[0] );	}
	}

    /*
    if  ( utilIndexSetContains( &(ars->arBoundaryChars), buf[pos]&0xff ) )
	{ (*pPos)++; return buf[pos]&0xff; }
    */

    return w[0];
    }

/************************************************************************/
/*									*/
/*  Read a string fragment inside a rule.				*/
/*									*/
/************************************************************************/

static int indAffixReadRuleString(	int *			pBufPos,
					const char * const 	buf,
					const AffixRules *	ars,
					iconv_t			icv,
					RuleString *		rs,
					int			uptoComma )
    {
    int		bufPos= *pBufPos;

    int		found;
    int		tryQuote;
    int		quoted= 0;

    const int	lower= 1;

    tryQuote= bufPos;
    found= affixChar( ars, buf, icv, lower, &tryQuote );
    if  ( found == -1 && buf[bufPos] == '"' )
	{ quoted= 1; bufPos++;	}

    while( buf[bufPos] )
	{
	if  ( rs->rsLength >= AFFMAXLENGTH )
	    {
	    LLDEB(rs->rsLength,AFFMAXLENGTH);
	    return -1;
	    }

	found= affixChar( ars, buf, icv, lower, &bufPos );
	if  ( found < 0 )
	    {
	    if  ( isascii( buf[bufPos] ) && ! isalpha( buf[bufPos] ) )
		{
		if  ( ! quoted		||
		      buf[bufPos] != '"'	||
		      buf[bufPos+1] != ','	)
		    { SLDEB((char *)buf,found); return -1; }
		}

	    found= buf[bufPos++];
	    }

	rs->rsUtf16[rs->rsLength++]= ucdToLower( found );

	if  ( uptoComma && buf[bufPos] == ',' )
	    {
	    bufPos++;
	    while( isspace(buf[bufPos]) )
		{ bufPos++;	}
	    break;
	    }

	while( isspace(buf[bufPos]) )
	    { bufPos++;	}
	}

    rs->rsUtf16[rs->rsLength]= 0;
    *pBufPos= bufPos;
    return 0;
    }

/************************************************************************/
/*									*/
/*  Read the fragment of the rule that must be replaced.		*/
/*									*/
/*  1)  If the mask length is shorter than the string to be replaced,	*/
/*	extend it to the length of the string.				*/
/*  2)  Require the characters that are to be replaced for the rule.	*/
/*									*/
/************************************************************************/

static int indAffixReadReplaceFrom(	int *			pBufPos,
					const char * const 	buf,
					AffixRule *		ar,
					iconv_t			icv,
					const AffixRules *	ars )
    {
    const int	uptoComma= 1;
    int		i;
    int		from= 0;

    if  ( indAffixReadRuleString( pBufPos, buf, ars, icv,
				    &(ar->arReplaceFrom), uptoComma ) )
	{ LDEB(1); return -1;	}

    /* 1 */
    if  ( ar->arReplaceFrom.rsLength > ar->arMaskLength )
	{
	int	extra= ar->arReplaceFrom.rsLength- ar->arMaskLength;
	int	i;

	if  ( ar->arWhatSide == 0 )
	    { LDEB(ar->arWhatSide);	}

	/*  suffix rule: Extend before the rule */
	if  ( ar->arWhatSide > 0 )
	    {
	    for ( i= ar->arReplaceFrom.rsLength- 1; i >= extra; i-- )
		{ ar->arMasks[i]= ar->arMasks[i- extra]; }

	    for ( i= 0; i < extra; i++ )
		{
		utilInitIndexSet( &( ar->arMasks[i].amSet) );
		ar->arMasks[i].amNegative= 1;
		}
	    }

	/*  prefix rule: Extend after the rule */
	if  ( ar->arWhatSide < 0 )
	    {
	    for ( i= ar->arMaskLength; i < ar->arReplaceFrom.rsLength; i++ )
		{
		utilInitIndexSet( &( ar->arMasks[i].amSet) );
		ar->arMasks[i].amNegative= 1;
		}
	    }

	ar->arMaskLength= ar->arReplaceFrom.rsLength;
	}

    /*  2  */
    if  ( ar->arWhatSide > 0 )
	{ from= ar->arMaskLength- ar->arReplaceFrom.rsLength;	}
    else{ from= 0;						}

    if  ( from < 0 )
	{ LDEB(from); return -1;	}

    for ( i= 0; i < ar->arReplaceFrom.rsLength; i++ )
	{
	ApplicabilityMask *	am= ar->arMasks+ i+ from;
	int			found= ar->arReplaceFrom.rsUtf16[i];

	if  ( am->amNegative )
	    {
	    if  ( utilIndexSetContains( &(am->amSet), found ) )
		{
		LXDEB(am->amNegative,found);
		LLDEB(i,from);
		indAffixListRule( -1, ar );
		return -1;
		}
	    }
	else{
	    if  ( ! utilIndexSetContains( &(am->amSet), found ) )
		{
		LXDEB(am->amNegative,found);
		LLDEB(i,from);
		indAffixListMask( am );
		indAffixListRule( -1, ar );
		return -1;
		}
	    }

	am->amNegative= 0;
	utilCleanIndexSet( &(am->amSet) );
	utilInitIndexSet( &(am->amSet) );
	if  ( utilIndexSetAdd( &(am->amSet), found ) )
	    { LDEB(found); return -1;	}
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Read the fragment of the rule that is the replacement.		*/
/*									*/
/************************************************************************/

static int indAffixReadReplaceWith(	int *			pBufPos,
					const char * const 	buf,
					AffixRule *		ar,
					iconv_t			icv,
					const AffixRules *	ars )
    {
    const int	uptoComma= 0;

    return indAffixReadRuleString( pBufPos, buf, ars, icv,
				    &(ar->arReplaceWith), uptoComma );
    }

/************************************************************************/
/*									*/
/*  Read the left hand side of a rule.					*/
/*									*/
/************************************************************************/

static int indAffixReadMask(		int *			pBufPos,
					const char * const 	buf,
					ApplicabilityMask *	am,
					iconv_t			icv,
					const AffixRules *	ars )
    {
    int		bufPos= *pBufPos;

    int		found;

    const int	lower= 1;

    am->amNegative= 0;
    utilCleanIndexSet( &(am->amSet) );
    utilInitIndexSet( &(am->amSet) );

    if  ( buf[bufPos] == '[' )
	{
	if  ( buf[bufPos+1] == '^' )
	    { am->amNegative= 1; bufPos += 2;	}
	else{ bufPos++;				}

	if  ( ! buf[bufPos] )
	    { SDEB((char *)buf); return -1; }

	while( buf[bufPos] != ']' )
	    {
	    found= affixChar( ars, buf, icv, lower, &bufPos );
	    if  ( found < 0 )
		{ SLDEB((char *)buf,found); return -1; }

	    if  ( utilIndexSetAdd( &(am->amSet), ucdToLower( found ) ) )
		{ LDEB(found); return -1;	}
	    }

	bufPos++; /* skip ']' */
	}
    else{
	if  ( ! buf[bufPos] )
	    { SDEB((char *)buf); return -1; }

	if  ( buf[bufPos] == '.' )
	    {
	    am->amNegative= 1; /* Not in the empty set: everything */
	    bufPos++;
	    }
	else{
	    found= affixChar( ars, buf, icv, lower, &bufPos );
	    if  ( found < 0 )
		{ SSLDEB((char *)buf+bufPos,(char *)buf,found); return -1; }

	    if  ( utilIndexSetAdd( &(am->amSet), ucdToLower( found ) ) )
		{ LDEB(found); return -1;	}
	    }
	}

    *pBufPos= bufPos;
    return 0;
    }

static int indAffixReadMasks(		int *			pBufPos,
					const char * const 	buf,
					AffixRule *		ar,
					iconv_t			icv,
					const AffixRules *	ars )
    {
    int		bufPos= *pBufPos;

    while( buf[bufPos] && buf[bufPos] != '>' )
	{
	if  ( ar->arMaskLength >= AFFMAXLENGTH )
	    { LLDEB(ar->arMaskLength,AFFMAXLENGTH); return -1; }

	if  ( indAffixReadMask( &bufPos, buf,
				    ar->arMasks+ ar->arMaskLength, icv, ars ) )
	    { LDEB(1); return -1;	}

	ar->arMaskLength++;

	while( isspace(buf[bufPos]) )
	    { bufPos++;	}
	}

    *pBufPos= bufPos;
    return 0;
    }

/************************************************************************/
/*									*/
/*  Read one affix rule.						*/
/*									*/
/************************************************************************/

static AffixRule * affixNewRule(	AffixRules *		ars,
					iconv_t			icv,
					const char *		file,
					int			line,
					int			whatSide,
					int			shift,
					int			cross,
					const char *		buf,
					AffixRule *		next )
    {
    int		bufPos= 0;
    AffixRule *	rval= (AffixRule *)0;

    if  ( ! whatSide )
	{ LDEB(whatSide); goto failure;	}

    rval= (AffixRule *)malloc( sizeof(AffixRule) );
    if  ( ! rval )
	{ LDEB(rval); goto failure;	}
    affixInitRule( rval );

    rval->arWhatSide= whatSide;
    rval->arShift= shift;
    rval->arCross= cross;
    rval->arNext= next;

    /*
    if  ( buf[bufPos] == '>' )
	{ SLSSDEB(file,line,(char *)buf,(char *)buf+bufPos);	}
    */

    rval->arMaskLength= 0;
    if  ( indAffixReadMasks( &bufPos, buf, rval, icv, ars ) )
	{ LDEB(1); goto failure;	}

    if  ( buf[bufPos] != '>' )
	{ SSDEB((char *)buf,(char *)buf+bufPos); goto failure; }

    bufPos++; /* skip '>' */

    while( isspace(buf[bufPos]) )
	{ bufPos++;	}

    if  ( ! buf[bufPos] )
	{ SLDEB((char *)buf,bufPos); goto failure; }

    rval->arReplaceFrom.rsLength= 0;

    if  ( buf[bufPos] == '-' )
	{
	bufPos++;

	if  ( indAffixReadReplaceFrom( &bufPos, buf, rval, icv, ars ) )
	    { LDEB(1); goto failure;	}
	}

    if  ( buf[bufPos] == '-' )
	{
	bufPos++;
	rval->arReplaceWith.rsLength= 0;
	}
    else{
	if  ( indAffixReadReplaceWith( &bufPos, buf, rval, icv, ars ) )
	    { LDEB(1); goto failure;	}
	}

    while( isspace(buf[bufPos]) )
	{ bufPos++;	}

    if  ( buf[bufPos] )
	{ SSDEB((char *)buf,(char *)buf+bufPos); goto failure; }

    /*
    fprintf( stderr, "MaskLength=%d, arFromLength= %d, rval->arToLength=%d\n",
	    rval->arMaskLength, rval->arFromLength, rval->arToLength );
    */

    return rval;

  failure:

    if  ( rval )
	{ free( rval );	}

    return (AffixRule *)0;
    }

/************************************************************************/
/*									*/
/*  Add a string char to the set of string chars.			*/
/*									*/
/************************************************************************/

#   define	AFFscSET(f,l,s2,m) \
		strcpy( ((char *)f->scString), (char *)(s2) ); \
		f->scLength= l; \
		f->scMeaningUtf16= m;

static int affixRememberStringChar(	AffixRules *	ars,
					const char *	s,
					int		l )
    {
    StringChar *	sc;

    sc= realloc( ars->arStringChars,
			(ars->arStringCharCount+ 1)* sizeof( StringChar ) );
    if  ( ! sc )
	{ LXDEB(ars->arStringCharCount,sc); return -1;	}

    ars->arStringChars= sc;
    sc += ars->arStringCharCount;

    if  ( l == 1 )
	{
	unsigned char scratch[2];
	scratch[0]= (*s)&0xff; scratch[1]= '\0';
	AFFscSET( sc, l, scratch, s[0]&0xff ); ars->arStringCharCount++; return 0;
	}

    /*  "  */
    if  ( l == 3 && ! strncmp( (char *)s, "a\\\"", l ) )
	{ AFFscSET( sc, l- 1, "a\"", ISO1_adieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "A\\\"", l ) )
	{ AFFscSET( sc, l- 1, "A\"", ISO1_Adieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "e\\\"", l ) )
	{ AFFscSET( sc, l- 1, "e\"", ISO1_edieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "E\\\"", l ) )
	{ AFFscSET( sc, l- 1, "E\"", ISO1_Edieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "i\\\"", l ) )
	{ AFFscSET( sc, l- 1, "i\"", ISO1_idieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "I\\\"", l ) )
	{ AFFscSET( sc, l- 1, "I\"", ISO1_Idieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "o\\\"", l ) )
	{ AFFscSET( sc, l- 1, "o\"", ISO1_odieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "O\\\"", l ) )
	{ AFFscSET( sc, l- 1, "O\"", ISO1_Odieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "u\\\"", l ) )
	{ AFFscSET( sc, l- 1, "u\"", ISO1_udieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "U\\\"", l ) )
	{ AFFscSET( sc, l- 1, "U\"", ISO1_Udieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "y\\\"", l ) )
	{ AFFscSET( sc, l- 1, "y\"", ISO1_ydieresis ); ars->arStringCharCount++; return 0; }

    /*  '  */
    if  ( l == 3 && ! strncmp( (char *)s, "a\\'", l ) )
	{ AFFscSET( sc, 2, "a'", ISO1_aacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "A\\'", l ) )
	{ AFFscSET( sc, 2, "A'", ISO1_Aacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "e\\'", l ) )
	{ AFFscSET( sc, 2, "e'", ISO1_eacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "E\\'", l ) )
	{ AFFscSET( sc, 2, "E'", ISO1_Eacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "i\\'", l ) )
	{ AFFscSET( sc, 2, "i'", ISO1_iacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "I\\'", l ) )
	{ AFFscSET( sc, 2, "I'", ISO1_Iacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "o\\'", l ) )
	{ AFFscSET( sc, 2, "o'", ISO1_oacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "O\\'", l ) )
	{ AFFscSET( sc, 2, "O'", ISO1_Oacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "u\\'", l ) )
	{ AFFscSET( sc, 2, "u'", ISO1_uacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "U\\'", l ) )
	{ AFFscSET( sc, 2, "U'", ISO1_Uacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "y\\'", l ) )
	{ AFFscSET( sc, 2, "y'", ISO1_yacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "Y\\'", l ) )
	{ AFFscSET( sc, 2, "Y'", ISO1_Yacute ); ars->arStringCharCount++; return 0; }

    /*  `  */
    if  ( l == 3 && ! strncmp( (char *)s, "a\\`", l ) )
	{ AFFscSET( sc, 2, "a`", ISO1_agrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "A\\`", l ) )
	{ AFFscSET( sc, 2, "A`", ISO1_Agrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "e\\`", l ) )
	{ AFFscSET( sc, 2, "e`", ISO1_egrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "E\\`", l ) )
	{ AFFscSET( sc, 2, "E`", ISO1_Egrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "i\\`", l ) )
	{ AFFscSET( sc, 2, "i`", ISO1_igrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "I\\`", l ) )
	{ AFFscSET( sc, 2, "I`", ISO1_Igrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "o\\`", l ) )
	{ AFFscSET( sc, 2, "o`", ISO1_ograve ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "O\\`", l ) )
	{ AFFscSET( sc, 2, "O`", ISO1_Ograve ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "u\\`", l ) )
	{ AFFscSET( sc, 2, "u`", ISO1_ugrave ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "U\\`", l ) )
	{ AFFscSET( sc, 2, "U`", ISO1_Ugrave ); ars->arStringCharCount++; return 0; }

    /*  ^  */
    if  ( l == 3 && ! strncmp( (char *)s, "a\\^", l ) )
	{ AFFscSET( sc, 2, "a^", ISO1_acircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "A\\^", l ) )
	{ AFFscSET( sc, 2, "A^", ISO1_Acircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "e\\^", l ) )
	{ AFFscSET( sc, 2, "e^", ISO1_ecircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "E\\^", l ) )
	{ AFFscSET( sc, 2, "E^", ISO1_Ecircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "i\\^", l ) )
	{ AFFscSET( sc, 2, "i^", ISO1_icircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "I\\^", l ) )
	{ AFFscSET( sc, 2, "I^", ISO1_Icircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "o\\^", l ) )
	{ AFFscSET( sc, 2, "o^", ISO1_ocircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "O\\^", l ) )
	{ AFFscSET( sc, 2, "O^", ISO1_Ocircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "u\\^", l ) )
	{ AFFscSET( sc, 2, "u^", ISO1_ucircumflex ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "U\\^", l ) )
	{ AFFscSET( sc, 2, "U^", ISO1_Ucircumflex ); ars->arStringCharCount++; return 0; }

    if  ( l == 2 && ! strncmp( (char *)s, "c,", l ) )
	{ AFFscSET( sc, l, "c,", ISO1_ccedilla ); ars->arStringCharCount++; return 0; }
    if  ( l == 2 && ! strncmp( (char *)s, "C,", l ) )
	{ AFFscSET( sc, l, "C,", ISO1_Ccedilla ); ars->arStringCharCount++; return 0; }

/* No!
    if  ( l == 2 && ! strncmp( (char *)s, "SS", l ) )
	{ AFFscSET( sc, l, "SS", ISO1_germandbls ); ars->arStringCharCount++; return 0; }
    if  ( l == 2 && ! strncmp( (char *)s, "ss", l ) )
	{ AFFscSET( sc, l, "ss", ISO1_germandbls ); ars->arStringCharCount++; return 0; }
*/
    if  ( l == 2 && ! strncmp( (char *)s, "SS", l ) )
	{ return 0; }
    if  ( l == 2 && ! strncmp( (char *)s, "ss", l ) )
	{ return 0; }

    if  ( l == 2 && ! strncmp( (char *)s, "sS", l ) )
	{ AFFscSET( sc, l, "sS", ISO1_germandbls ); ars->arStringCharCount++; return 0; }
    if  ( l == 2 && ! strncmp( (char *)s, "Ss", l ) )
	{ AFFscSET( sc, l, "Ss", ISO1_germandbls ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'a", l ) )
	{ AFFscSET( sc, 2, "'a", ISO1_aacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'A", l ) )
	{ AFFscSET( sc, 2, "'A", ISO1_Aacute ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'e", l ) )
	{ AFFscSET( sc, 2, "'e", ISO1_eacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'E", l ) )
	{ AFFscSET( sc, 2, "'E", ISO1_Eacute ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'i", l ) )
	{ AFFscSET( sc, 2, "'i", ISO1_iacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'I", l ) )
	{ AFFscSET( sc, 2, "'I", ISO1_Iacute ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'o", l ) )
	{ AFFscSET( sc, 2, "'o", ISO1_oacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'O", l ) )
	{ AFFscSET( sc, 2, "'O", ISO1_Oacute ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'u", l ) )
	{ AFFscSET( sc, 2, "'u", ISO1_uacute ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'U", l ) )
	{ AFFscSET( sc, 2, "'U", ISO1_Uacute ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\'n", l ) )
	{ AFFscSET( sc, 2, "'n", ISO1_ntilde ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\'N", l ) )
	{ AFFscSET( sc, 2, "'N", ISO1_Ntilde ); ars->arStringCharCount++; return 0; }

    if  ( l == 3 && ! strncmp( (char *)s, "\\\"u", l ) )
	{ AFFscSET( sc, 2, "\"u", ISO1_udieresis ); ars->arStringCharCount++; return 0; }
    if  ( l == 3 && ! strncmp( (char *)s, "\\\"U", l ) )
	{ AFFscSET( sc, 2, "\"U", ISO1_Udieresis ); ars->arStringCharCount++; return 0; }

    SLDEB((char *)s,l); return -1;
    }

/************************************************************************/
/*									*/
/*  Remember string chars.						*/
/*									*/
/************************************************************************/

static int	affixRememberStringChars(	AffixRules *	ars,
						const char *	buf,
						int *		pPos )
    {
    int		pos= *pPos;

    while( buf[pos] == ' ' || buf[pos] == '\t' )
	{ pos++;	}

    while( buf[pos] )
	{
	int	before= pos;

	while( buf[pos] && buf[pos] != ' ' && buf[pos] != '\t' )
	    { pos++;	}

	if  ( affixRememberStringChar( ars, buf+ before, pos- before ) )
	    { SDEB((char *)buf); return -1;	}

	while( buf[pos] == ' ' || buf[pos] == '\t' )
	    { pos++;	}
	}

    *pPos= pos; return 0;
    }

/************************************************************************/
/*									*/
/*  Fill a character set.						*/
/*									*/
/************************************************************************/

static int	affixGetChars(	IndexSet *		is,
				const char *		buf,
				int *			pPos )
    {
    int		prev= -1;
    int		pos= *pPos;

    while( buf[pos] == ' ' || buf[pos] == '\t' )
	{ pos++;	}

    while( buf[pos] )
	{
	switch( buf[pos] )
	    {
	    unsigned int		cc;

	    case '"':
		pos++;
		if  ( utilIndexSetAdd( is, buf[pos]&0xff ) )
		    { LDEB( buf[pos] ); return -1;	}
		pos++;
		if  ( buf[pos] != '"' )
		    { SSDEB((char *)buf,(char *)buf+pos); return -1;	}
		pos++; break;

	    case '[':
		pos++;
		while( buf[pos] != ']' )
		    {
		    if  ( ! buf[pos] )
			{ SLDEB((char *)buf,pos); return -1;	}

		    if  ( prev >= 0 && buf[pos] == '-' )
			{
			if  ( buf[pos+1] == '\\' && buf[pos+2] == 'x' )
			    {
			    sscanf( (const char *)buf+ pos+ 3, "%2x", &cc );

			    while( prev < cc )
				{
				if  ( utilIndexSetAdd( is, prev ) )
				    { LDEB( prev ); return -1;	}
				prev++;
				}

			    if  ( utilIndexSetAdd( is, cc ) )
				{ LDEB( cc ); return -1;	}
			    prev= -1; pos += 5;
			    continue;
			    }
			else{
			    while( prev < buf[pos+1] )
				{
				if  ( utilIndexSetAdd( is, prev ) )
				    { LDEB( prev ); return -1;	}
				prev++;
				}

			    if  ( utilIndexSetAdd( is, buf[pos+1]&0xff ) )
				{ LDEB( buf[pos+1] ); return -1;	}
			    prev= -1; pos += 2;
			    continue;
			    }
			}

		    if  ( buf[pos] == '\\' && buf[pos+1] == 'x' )
			{
			sscanf( (const char *)buf+ pos+ 2, "%2x", &cc );

			if  ( utilIndexSetAdd( is, cc ) )
			    { LDEB( cc ); return -1;	}
			prev= cc; pos += 4;
			continue;
			}

		    if  ( buf[pos] == '\\' && buf[pos+1] == '-' )
			{ pos++; }

		    if  ( utilIndexSetAdd( is, buf[pos]&0xff ) )
			{ LDEB( buf[pos] ); return -1;	}
		    prev= buf[pos]&0xff; pos++;
		    continue;
		    }
		prev= -1; pos++; break;
	    case '\\':
		if  ( buf[pos+1] == 'x' )
		    {
		    sscanf( (const char *)buf+ pos+ 2, "%2x", &cc );
		    if  ( utilIndexSetAdd( is, cc ) )
			{ LDEB( cc ); return -1;	}
		    prev= cc; pos += 4;
		    break;
		    }

		if  ( utilIndexSetAdd( is, buf[pos+ 1]&0xff ) )
		    { LDEB( buf[pos] ); return -1;	}
		pos += 2; break;

	    default:
		if  ( utilIndexSetAdd( is, buf[pos]&0xff ) )
		    { LDEB( buf[pos] ); return -1;	}
		pos++; break;
	    }

	if  ( buf[pos] && buf[pos] != ' ' && buf[pos] != '\t' )
	    { SSDEB((char *)buf,(char *)buf+pos); return -1;	}

	while( buf[pos] == ' ' || buf[pos] == '\t' )
	    { pos++;	}
	}

    *pPos= pos; return 0;
    }

/************************************************************************/
/*									*/
/*  Read affix file.							*/
/*									*/
/************************************************************************/

AffixRules * indAffixReadFile(	const char *		filename,
				const AffixThrough *	at )
    {
    FILE *		f= fopen( filename, "r" );
    int			pos;
    int			words;
    int			whatSide= 0;
    int			shift= 0;
    int			cross= 0;
    int			compound= 0;
    int			currentFlag= -1;
    int			l;
    int			line= 0;

    char		buf[BUL+1];
    char		word1[BUL+1];
    char		word2[BUL+1];

    AffixRule *	newRule;

    AffixRules * 	ars= (AffixRules *)0;

    if  ( ! f )
	{ SXDEB(filename,f); goto failure;	}

    ars= malloc( sizeof(AffixRules) );
    if  ( ! ars )
	{ XDEB(ars); goto failure;	}

    affixInitRules( ars );

    while( fgets( buf, BUL, f ) )
	{
	line++;

	if  ( ! strcmp( (char *)buf, "#up\n" ) )
	    { shift=  1; continue;	}
	if  ( ! strcmp( (char *)buf, "#down\n" ) )
	    { shift= -1; continue;	}

	l= 0;
	while( l < BUL && buf[l] != '#' && buf[l] != '\n' )
	    { l++;	}
	buf[l]= '\0';

	pos= 0;
	while( buf[pos] == ' ' || buf[pos] == '\t' )
	    { pos++;	}

	if  ( ! buf[pos] )
	    { continue;	}

	words= 0;
	{
	int	b= 0;
	while( buf[b] && isalpha( buf[b] ) )
	    { b++;	}

	if  ( b > 0 )
	    {
	    int		b2;

	    strncpy( word1, buf, b )[b]= '\0';
	    words++;

	    while( buf[b] && isspace( buf[b] ) )
		{ b++;	}

	    b2= 0;
	    while( buf[b+ b2] && ! isspace( buf[b+ b2] ) )
		{ b2++;	}

	    if  ( b2 > 0 )
		{
		strncpy( word2, buf+ b, b2 )[b2]= '\0';
		words++;
		}
	    }
	}

	if  ( words > 0 )
	    {
	    if  ( ! strcmp( (char *)word1, "flagmarker" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "defstringtype" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "compoundwords" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "allaffixes" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "altstringtype" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "altstringchar" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "texchars" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "nroffchars" )	)
		{ continue;	}
	    if  ( ! strcmp( (char *)word1, "compoundmin" )	)
		{ continue;	}

	    if  ( ! strcmp( (char *)word1, "stringchar" )	)
		{
		l= 10;
		if  ( affixRememberStringChars( ars, buf, &l ) )
		    { SLSDEB(filename,line,(char *)buf); goto failure;	}
		continue;
		}

	    if  ( ! strcmp( (char *)word1, "boundarychars" )	)
		{
		l= 13;
		if  ( affixGetChars( &(ars->arBoundaryChars), buf, &l ) )
		    { SLSDEB(filename,line,(char *)buf); goto failure;	}
		continue;
		}

	    if  ( ! strcmp( (char *)word1, "wordchars" )	)
		{
		l= 9;
		if  ( affixGetChars( &(ars->arWordChars), buf, &l ) )
		    { SLSDEB(filename,line,(char *)buf); goto failure;	}
		continue;
		}

	    if  ( ! strcmp( (char *)word1, "prefixes" )	)
		{ whatSide= -1; continue;	}
	    if  ( ! strcmp( (char *)word1, "suffixes" )	)
		{ whatSide=  1; continue;	}

	    if  ( ! strcmp( (char *)word1, "flag" )	)
		{
		shift= 0; cross= 0; compound= 0;

		if  ( words > 1 )
		    {
		    char *	w2= word2;

		    if  ( w2[0] == '\\' )
			{ w2++;	}

		    if  ( w2[0] == '*' )
			{
			if  ( w2[1] == '\\' )
			    { w2++;	}

			if  ( strlen( (char *)w2 ) == 3 && w2[2] == ':' )
			    { cross= 1; currentFlag= w2[1]; continue; }
			}
		    else{
			if  ( strlen( (char *)w2 ) == 2 && w2[1] == ':' )
			    { cross= 0; currentFlag= w2[0]; continue; }
			}

		    if  ( w2[0] == '~' )
			{
			if  ( w2[1] == '\\' )
			    { w2++;	}

			if  ( strlen( (char *)w2 ) == 3 && w2[2] == ':' )
			    { compound= 1; currentFlag= w2[1]; continue; }
			}
		    else{
			if  ( strlen( (char *)w2 ) == 2 && w2[1] == ':' )
			    { compound= 0; currentFlag= w2[0]; continue; }
			}
		    }
		}
	    }

	/* Incomplete line ? */
	while( ! strchr( (char *)buf, '>' ) && l < BUL )
	    {
	    if  ( ! fgets( (char *)buf+ l, BUL- l, f ) )
		{ SLDEB(filename,line); goto failure;	}

	    line++;

	    while( l < BUL && buf[l] != '#' && buf[l] != '\n' )
		{ l++;	}

	    if  ( buf[l] == '#' )
		{ buf[l]= '\0'; break; }
	    buf[l]= '\0';
	    }

	if  ( ! compound )
	    {
	    newRule= affixNewRule( ars, at->atIconv, filename, line,
				    whatSide, shift, cross,
				    buf+ pos, ars->arsRules[currentFlag] );
	    if  ( ! newRule )
		{ SLDEB(filename,line); SDEB((char *)buf); goto failure; }

	    ars->arsRules[currentFlag]= newRule;
	    }
	}

    fclose( f ); return ars;

  failure:

    if  ( f )
	{ fclose( f );	}
    if  ( ars )
	{ affixFreeRules( ars );	}

    return (void *)0;
    }

